import requests, os, getpass
from http import HTTPStatus
from . import HeaderType, RequestContentType, ResponseContentType, StoreType

class ClientForm:
    r""" Represents a form with fields and attachements that can sent to server using a POST request 
    
    example:

        form = ClientForm(
            name = "my_name",
            age = "100", 
        ).attach(
            alias="friendly_name1", 
            name="original_name1", 
            mime="file_type_hint", 
            handle=<name or handle or buffer>,
        ).attach(...).attach(...)

        the arguments to `__init__` should be string only - other data-type is not supported (it uses forms in POST method)
        the `attach` function returns the object itself so can be chained

        Note - handle can be either one of:
        # file name as string,  `f="/home/user/abc.txt"` - will open the file and close it after sending
        # open file handle,     `f=open("/home/user/abc.txt", 'r')` - will be automatically closed after sending
        # bytes buffer,         `f=io.BytesIO(); f.write(<something>);` - will seek to position-0 and close after sending
    """

    def __init__(self, **kwargs):
        self.data = {f'{k}':f'{v}' for k,v in kwargs.items()}
        self.attached={}
        self.files={}
    
    def attach(self, alias, name, mime, handle): 
        # handle can be either a file-path or a BytesIO object
        self.attached[alias] = dict(name=name, handle=handle, mime=mime, htype=isinstance(handle, str))
        return self

    def clear(self, data=True, files=True):
        if data: self.data.clear()
        if files: self.files.clear()

    def open(self):
        self.files.clear()
        for alias,info in self.attached.items():
            try:
                handle = open(info['handle'], 'rb') if info['htype'] else info['handle']
                handle.seek(0)
                self.files[alias] = (info['name'], handle, info['mime'])
            except: pass

    def close(self):
        for _, h, _ in self.files.values(): h.close()

class ClientResponse:
    r""" represents the response from a `send` request """
    truesym = '✔️'
    falsesym = '❌'
    def __init__(self, status_ok, xtype, xtag, xresponse):
        self.ok = status_ok
        
        self.type = xtype
        self.tag = xtag
        self.response = xresponse

    def __bool__(self): return self.ok
    def __str__(self): return f'[{self.truesym if self.ok else self.falsesym}] | Type={self.type} | Tag={self.tag} | Content={type(self.response)}'
    def __call__(self): return self.response

class Client:
    r""" HTTP Client Class - Represents a client that will access the API 
    The main methods to send data are:
    `send_` methods 
    `path_` methods 
    both type of methods return a `ClientResponse` object
    """

    def __init__(self, uid:str='', server:str='localhost:8080'):
        self.uid = uid if uid else getpass.getuser()
        assert len(self.uid)>0, f'Invalid uid {type(self.uid)}::[{self.uid}]'
        self.server = server
        self.url = f'http://{self.server}/'
        self.store = f'http://{self.server}/store/'
        self.timeout = None # (float or tuple) – How many seconds to wait for the server to send data - can be (connect timeout, read timeout) tuple.
        self.allow_redirects = False # we keep this False, only server will respond
        #self.params = None  # this is added to url, so make sure to pass strings only - both keys and values

    def __str__(self): return self.uid

    def check(self): # verify connection 
        # make a simple get request - the api should respond with ok
        try:        is_ok = requests.get(self.url, timeout=self.timeout).ok 
        except:     is_ok = False
        return      is_ok

    def send(self, xcontent:object, xtype:RequestContentType,  xstream:bool=False):
        # xtype is RequestContentType
        if xtype==RequestContentType.MESG: 
            xjson, xdata, xfiles = None, f'{xcontent}'.format('utf-8'), None
        elif xtype==RequestContentType.BYTE: 
            assert type(xcontent) is bytes, f'Expecting bytes but got {type(xcontent)}'
            xjson, xdata, xfiles = None, xcontent, None
        elif xtype==RequestContentType.FORM: 
            assert type(xcontent) is ClientForm
            xjson, xdata, xfiles = None, xcontent.data, xcontent.files
            xcontent.open()
        elif xtype==RequestContentType.JSON: 
            xjson, xdata, xfiles = xcontent, None, None
        else:               
            raise TypeError(f'Type "{xtype}" is not a valid content type') # xtype must be in ClientContentType

        # make a request to server ClientResponse
        try:
            response = requests.post(
                url=            self.url, allow_redirects=self.allow_redirects,  timeout=self.timeout,  #params=self.params,
                headers=        {HeaderType.XTYPE:xtype, HeaderType.XTAG:self.uid},
                stream=         xstream,
                json=           xjson,
                data=           xdata,
                files=          xfiles,
            )
        except: response = None
        finally:
            if xtype==RequestContentType.FORM: xcontent.close()
        return self.handle_response(response, xstream)

    def handle_response(self, response, xstream):
        # handle the response
        # NOTE: the `response` object contains the `request` object that we sent in response.request 
        # headers are sent always (independent of stream=True/False)
        if response is None:  status_ok, xtype, xtag, xresponse = False, None, None, None
        else:
            status_code =   response.status_code
            status_ok =     response.ok
            xtag =          response.headers.get(HeaderType.XTAG)
            xtype =         response.headers.get(HeaderType.XTYPE) # response content type

            if   status_code == HTTPStatus.OK: 
                if   xtype==ResponseContentType.MESG:       xresponse = response.content.decode('utf-8')
                elif xtype==ResponseContentType.BYTE:       xresponse = response.content
                elif xtype==ResponseContentType.JSON:       xresponse = response.json()
                else:                                       xresponse = None      
            elif status_code == HTTPStatus.NOT_ACCEPTABLE:  xresponse = None  
            elif status_code == HTTPStatus.NOT_FOUND:       xresponse = None  
            else:                                           xresponse = None   # this should not happen
            #if xstream: pass
            #else:        pass
            response.close()
        return ClientResponse(status_ok, xtype, xtag, xresponse)

    """ send_ methods """

    def send_mesg(self, message:str):               return self.send(f'{message}',  RequestContentType.MESG, xstream=False )
    def send_json(self, json_object:object):        return self.send(json_object,   RequestContentType.JSON, xstream=False )
    def send_form(self, client_form:ClientForm):    return self.send(client_form,   RequestContentType.FORM, xstream=False )
    def send_byte(self, byte_data:bytes):           return self.send(byte_data,     RequestContentType.BYTE, xstream=False )

    """ path_ methods """

    def path_get(self, path=None, save=None):
        r""" Query the store to get files and folders 
        
        `path`:         <str> the path on the server to get from. 
                        If path is a file, it will download the file and save it at the path provided in header `XTAG` (it provides a filename)
                        If path is a folder, it will return a dict of listing dict(root=?, files=?, folders=?) `localhost:8080/store/path/to/folder`
                        if path is empty string "", gets the listing from root folder `localhost:8080/store/`
                        If path is None, does a directory listing at top level `localhost:8080/store`
                        
        `save`:        <str> (optional) the local path to save an incoming file, 
                        If None, uses the header `XTAG` (not required for listing directory - only for file get)

        """
        try:    response = requests.get( url = ( self.store[:-1] if path is None else os.path.join(self.store, path) ), timeout = self.timeout, headers={HeaderType.XTAG: self.uid})
        except: return ClientResponse(False, None, None, "Exception in path_get.get method")
        uok = response.ok 
        utype = response.headers.get(HeaderType.XTYPE)
        utag = response.headers.get(HeaderType.XTAG)
        ures=None
        if uok:
            if  utype ==  StoreType.HOME or utype == StoreType.DIR: ures = response.json()
            elif utype == StoreType.FILE: 
                if not save: save = utag
                if not save: ures = f"Invalid save location [{save}]"
                else:
                    try:
                        with open(save, 'wb') as f: f.write(response.content)
                    except: ures = f"Error Saving incoming file at {save}"
            else:       ures = f"Response type {utype} is unexpected for this request"
        else:           ures = f"Response not ok \n\n{response.text}" 
        response.close()
        return ClientResponse(uok, utype, utag, ures)
    def path_set(self, path, item=None):
        r""" Put files and folders on the server
        
        `path`:         <str> the path on the server to set at. 
        `item`:         the local path of a file to send (only when sending files not folders)
                        If item is a file, it will create a file on the server at `path` (stream file to server)
                        If item is None, it will create a folder at `path`
                        if item is anything else, error will be thrown

        """
        if item is None: 
            try:response = requests.put(url=os.path.join(self.store, path), timeout=self.timeout, headers={HeaderType.XTAG: self.uid})
            except: return ClientResponse(False, None, None, "Exception in path_set.put method")
        elif os.path.isfile(item):
            try:
                with open(item, 'rb') as f:
                    response = requests.post(url=os.path.join(self.store, path), data=f, timeout=self.timeout, headers={HeaderType.XTAG: self.uid})
            except: return ClientResponse(False, None, None, "Exception in path_set.post method")
        else: raise FileNotFoundError(f'cannot find path {item}')
        uok = response.ok
        utype = response.headers.get(HeaderType.XTYPE)
        utag = response.headers.get(HeaderType.XTAG)
        ures = ""
        if uok:
            if utype ==  StoreType.MSG: ures = f'{response.text}'
            else:                       ures = f"Response type {utype} is unexpected for this request"
        else:                           ures = f"Response not ok \n\n{response.text}"  
        response.close()
        return ClientResponse(uok, utype, utag, ures)
    def path_del(self, path, recursive=False):
        r""" Delete files and folders from the server
        
        `path`:         <str> the path on the server to delete. 
                        If path is a file on the server, it will be deleted
                        If path is a folder on the server, it will be deleted only if its empty (set recurvie=True for recursive delete)
        """
        # only this request uses the XTYPE header to indicate if directory has to be recurviely deleted or not
        try: response = requests.delete(url= os.path.join(self.store, path), timeout=self.timeout, headers={HeaderType.XTAG: self.uid, HeaderType.XTYPE: f'{int(recursive)}'})
        except: return ClientResponse(False, None, None, "Exception in path_del.delete method")
        uok = response.ok
        utype = response.headers.get(HeaderType.XTYPE)
        utag = response.headers.get(HeaderType.XTAG)
        ures = ""
        if uok:
            if utype ==  StoreType.MSG: ures = f'{response.text}'
            else:                       ures = f"Response type {utype} is unexpected for this request"
        else:                           ures = f"Response not ok \n\n{response.text}" 
        response.close()
        return ClientResponse(uok, utype, utag, ures)





__doc__=f""" 
-------------------------------------------------------------
Flask-based web api 
-------------------------------------------------------------
python -m pip install Flask waitress requests
python -m known.api --dir=/home/user --config=api
"""

DefaultArgs = {
    'dir'     : dict(   type=str, default='',    help="path of workspace directory"   ),
    'config'  : dict(   type=str, default='api', help="name of config-file"           ),
}

class HeaderType: # lists the http headers used to carry special info
    # headers are interpreted different 
    #   based on `request` and `response` object
    #   based on `api` and `store` access
    XTAG =      'known-api-tag'         # "Used to specify a Tag"
    XTYPE =     'known-api-type'        # "Used to specify a ContentType"


class RequestContentType: # lists the valid type of requests that clients can send back (specified in HeaderType.XTYPE header field)
    MESG = "MESG" # represents a string
    BYTE = "BYTE" # a stream of bytes
    JSON = "JSON" # a json serializable object
    FORM = "FORM" # a ClientForm with fields and attachements
    # only use either one of data or json in post request (not both)
    # form can only be sent from client to server but not other way

class ResponseContentType: # lists the valid type of response that server can send back to a `send` request (specified in HeaderType.XTYPE header field)
    MESG = "MESG" # represents a string
    BYTE = "BYTE" # a stream of bytes
    JSON = "JSON" # a json serializable object
    ALL = {str:MESG, dict:JSON, list:JSON, tuple:JSON, bytes:BYTE}
    #ALL = set(list(MAP.keys()))

class StoreType: # lists the valid type of response that server can send back to a `path` reuest (specified in HeaderType.XTYPE header field)
    HOME = "H" # Home view which lists all possible paths in the root directory as a dict of {path:files}
    DIR =  "D" # Directory view returns a json dict as dict(base=<base_path_of_dir>, files=<list_of_files>, folders=<list_of_folders>)
    FILE = "F" # Indicates a files on the server, this must be saved to a location on client
    MSG = "M"  # Indicates that this is a message




# Allow only execution - no importing
#-----------------------------------------------------------------------------------------
from sys import exit
if __name__!='__main__': exit(f'[!] can not import {__name__}.{__file__}')
#-----------------------------------------------------------------------------------------






#%% Imports
from . import Table, ZipFolders
import os, re, argparse, getpass, random, logging
#from math import inf
import datetime 
def fnow(format): return datetime.datetime.strftime(datetime.datetime.now(), format)
try:
    from flask import Flask, render_template, request, redirect, url_for, session, abort, send_file
    from flask_wtf import FlaskForm
    from wtforms import SubmitField, MultipleFileField
    from werkzeug.utils import secure_filename
    from wtforms.validators import InputRequired
    from waitress import serve
except: exit(f'[!] The required Flask packages missing:\tFlask>=3.0.2, Flask-WTF>=1.2.1\twaitress>=3.0.0\n  ⇒ pip install Flask Flask-WTF waitress')






#%% Parse Args

parser = argparse.ArgumentParser()
parser.add_argument('--title', type=str, default='Store',        help="a title to display at the login")
parser.add_argument('--root', type=str, default='.',             help="path of root directory")
parser.add_argument('--login', type=str, default='.login',       help="path of login file")
parser.add_argument('--secret', type=str, default='.secret',     help="path of secret file")

parser.add_argument('--ext', type=str, default='',               help="csv string of allowed extensions (excluding .)")
parser.add_argument('--mus', type=str, default='1024TB',         help="Maximum size of uploadable file (hhtp_body_size)")
parser.add_argument('--msl', type=int, default=100,              help="Maximum string length of auth policy")
parser.add_argument('--case', type=int, default=0,               help="case sensetive login uid")
parser.add_argument('--hide', type=int, default=0,               help="hide hidden files")

parser.add_argument('--host', type=str, default='0.0.0.0',       help="ip addr 0.0.0.0 for all")
parser.add_argument('--port', type=str, default='8080',          help="port")
parser.add_argument('--mcon', type=int, default=50,              help="max-connections")
parser.add_argument('--threads', type=int, default=4,            help="threads used by waitress")

parser.add_argument('--log', type=str, default='',               help="path of log file")
parser.add_argument('--verbose', type=int, default=2,            help="verbose level in logging")
parsed = parser.parse_args()








#%% Setup logging

# ------------------------------------------------------------------------------------------
LOGFILE = f'{parsed.log}'                               # define log dir - contains all logs
if LOGFILE and parsed.verbose>0: 
# ------------------------------------------------------------------------------------------
    try:
        LOGFILE = os.path.abspath(LOGFILE)
        logging.basicConfig(filename=LOGFILE, level=logging.INFO, format='%(asctime)s - %(message)s')
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(message)s')
        console_handler.setFormatter(formatter)
        logger = logging.getLogger()
        logger.addHandler(console_handler)
    except: exit(f'[!] Logging could not be setup at {LOGFILE}')
else: LOGFILE = None
# ------------------------------------------------------------------------------------------
if parsed.verbose==0: # no log
    def sprint(msg): pass
    def dprint(msg): pass
    def fexit(msg): exit(msg)
elif parsed.verbose==1: # only server logs
    if LOGFILE is None:
        def sprint(msg): print(msg) 
        def dprint(msg): pass 
        def fexit(msg): exit(msg)
    else:
        def sprint(msg): logging.info(msg) 
        def dprint(msg): pass 
        def fexit(msg):
            logging.error(msg) 
            exit()
elif parsed.verbose>=2: # server and user logs
    if LOGFILE is None:
        def sprint(msg): print(msg) 
        def dprint(msg): print(msg) 
        def fexit(msg): exit(msg)
    else:
        def sprint(msg): logging.info(msg) 
        def dprint(msg): logging.info(msg) 
        def fexit(msg):
            logging.error(msg) 
            exit()
else: raise ZeroDivisionError # impossible

sprint(f'Starting...')
sprint(f'↪ Logging @ {LOGFILE}')








#%% setup w

PYDIR = os.path.abspath(os.path.dirname(__file__))       # script directory of __main__.py
CWDIR = os.path.abspath(os.getcwd())    # location of run

# ------------------------------------------------------------------------------------------
ROOTDIR = f'{parsed.root}'
if not ROOTDIR: fexit(f'[!] Root directory was not specified')
ROOTDIR = os.path.abspath(ROOTDIR)
try: os.makedirs(ROOTDIR, exist_ok=True)
except: fexit(f'[!] Root directory was not found and could not be created')
sprint(f'↪ Root directory is {ROOTDIR}')

#-----------------------------------------------------------------------------------------

# ------------------------------------------------------------------------------------------
LOGINFILE = f'{parsed.login}' 
if not LOGINFILE: fexit(f'[!] Login-file was not specified')
LOGINFILE = os.path.abspath(LOGINFILE)

sprint(f'↪ Login-file is {LOGINFILE}')
db = Table.Create(columns=('access', 'uid', 'pass'), primary_key='uid')

if not os.path.isfile(LOGINFILE):
    if not os.path.exists(LOGINFILE):
        try: 
            THIS_USER = getpass.getuser()
            db[...] = ('UDXR', THIS_USER, '')
            db > LOGINFILE
            sprint(f'↪ New Login-file created @ {LOGINFILE} with user {THIS_USER}')
        except: fexit(f'[!] Login-file was not found and could not be created')
    else: fexit(f'[!] Login-file doesnt exists as a file')

try: 
    db < LOGINFILE
    sprint(f'⇒ Loaded Login-file: {LOGINFILE}')
except : fexit(f'[!] Login-file could not be loaded')
#-----------------------------------------------------------------------------------------

def NEW_SECRET_KEY():
    randx = lambda : random.randint(1111111111, 9999999999)
    r1 = randx()
    for _ in range(datetime.datetime.now().second): _ = randx()
    r2 = randx()
    for _ in range(datetime.datetime.now().second): _ = randx()
    r3 = randx()
    for _ in range(datetime.datetime.now().second): _ = randx()
    r4 = randx()
    return ':{}:{}:{}:{}:{}:'.format(r1,r2,r3,r4,fnow("%Y%m%d%H%M%S"))
# ------------------------------------------------------------------------------------------
SECRETFILE = f'{parsed.secret}' 
if not SECRETFILE: fexit(f'[!] Secret-file was not specified')
SECRETFILE = os.path.abspath(SECRETFILE)

sprint(f'↪ Secret-file is {SECRETFILE}')
if not os.path.isfile(SECRETFILE): #< --- if key dont exist, create it
    try:
        with open(SECRETFILE, 'w') as f: f.write( NEW_SECRET_KEY() ) #<---- auto-generated key
    except: fexit(f'[!] could not create secret key @ {SECRETFILE}')
    sprint(f'⇒ New secret created: {SECRETFILE}')

try:
    with open(SECRETFILE, 'r') as f: APP_SECRET_KEY = f.read()
    sprint(f'⇒ Loaded secret file: {SECRETFILE}')
except: fexit(f'[!] could not read secret key @ {SECRETFILE}')

# ------------------------------------------------------------------------------------------
#-----------------------------------------------------------------------------------------
#-----------------------------------------------------------------------------------------

#-----------------------------------------------------------------------------------------
# password policy
#-----------------------------------------------------------------------------------------
class AuthPolicy:

    def __init__(self, max_len) -> None:
        self.MAX_STR_LEN = int(max_len)

    def VALIDATE_PASS(self, instr):   # a function that can validate the password - returns bool type
        try: assert (len(instr) < self.MAX_STR_LEN) and bool(re.fullmatch("(\w|@|\.)+", instr)) # alpha_numeric @.
        except AssertionError: return False
        return True
    #-----------------------------------------------------------------------------------------
    # uid policy
    def VALIDATE_UID(self, instr):   # a function that can validate the uid - returns bool type
        try: assert (len(instr) < self.MAX_STR_LEN) and bool(re.fullmatch("(\w)+", instr)) # alpha_numeric 
        except AssertionError: return False
        return True
    #-----------------------------------------------------------------------------------------
    #-----------------------------------------------------------------------------------------
    # name policy
    def VALIDATE_NAME(self, instr): return  (len(instr) >0) and (len(instr) < self.MAX_STR_LEN) and \
        bool(re.fullmatch("((\w)(\w|\s)*(\w))|(\w)", instr)) # alpha-neumeric but no illegal spaces before or after
    #-----------------------------------------------------------------------------------------

class FilePolicy:

    @staticmethod
    def GET_VALID_RE_PATTERN(validext):
        if not validext: return ".+"
        pattern=""
        for e in validext: pattern+=f'{e}|'
        return pattern[:-1]

    def __init__(self, ext, mus) -> None:
        self.ALLOWED_EXTENSIONS = set([x.strip() for x in ext.split(',') if x])  # a set or list of file extensions that are allowed to be uploaded 
        if '' in self.ALLOWED_EXTENSIONS: self.ALLOWED_EXTENSIONS.remove('')
        self.VALID_FILES_PATTERN = self.GET_VALID_RE_PATTERN(self.ALLOWED_EXTENSIONS)
        self.MAX_UPLOAD_SIZE = __class__.str2bytes(mus) 
        self.MAX_UPLOAD_SIZE_HR = __class__.DISPLAY_SIZE_READABLE(self.MAX_UPLOAD_SIZE)

    #def __call__(self): return self.ALLOWED_EXTENSIONS, self.MAX_UPLOAD_SIZE_HR

    def __call__(self, filename):   # a function that checks for valid file extensions based on ALLOWED_EXTENSIONS
        if '.' in filename: 
            name, ext = filename.rsplit('.', 1)
            safename = f'{name}.{ext.lower()}'
            isvalid = bool(re.fullmatch(f'.+\.({self.VALID_FILES_PATTERN})$', safename))
        else:               
            name, ext = filename, ''
            safename = f'{name}'
            isvalid = (not self.ALLOWED_EXTENSIONS)
        return isvalid, safename

    @staticmethod
    def str2bytes(size):
        sizes = dict(KB=2**10, MB=2**20, GB=2**30, TB=2**40)
        return int(float(size[:-2])*sizes.get(size[-2:].upper(), 0))

    @staticmethod
    def DISPLAY_SIZE_READABLE(mus):
        # find max upload size in appropiate units
        mus_kb = mus/(2**10)
        if len(f'{int(mus_kb)}') < 4:
            mus_display = f'{mus_kb:.2f} KB'
        else:
            mus_mb = mus/(2**20)
            if len(f'{int(mus_mb)}') < 4:
                mus_display = f'{mus_mb:.2f} MB'
            else:
                mus_gb = mus/(2**30)
                if len(f'{int(mus_gb)}') < 4:
                    mus_display = f'{mus_gb:.2f} GB'
                else:
                    mus_tb = mus/(2**40)
                    mus_display = f'{mus_tb:.2f} TB'
        return mus_display


AP = AuthPolicy(parsed.msl)

FP = FilePolicy(parsed.ext, parsed.mus)

# print(f'{LOGFILE=}', f'{PYDIR=}', f'{CWDIR=}', f'{ROOTDIR=}', f'{LOGINFILE=}', f'{SECRETFILE=}')


#%%






















# ------------------------------------------------------------------------------------------
# html pages
# ------------------------------------------------------------------------------------------
#-----------------------------------------------------------------------------------------
# Create HTML
# ------------------------------------------------------------------------------------------

# ******************************************************************************************
CSS_TEMPLATES = dict(
# ****************************************************************************************** 0b7daa
style = """

.title{
    color: #000000;
    font-size: xxx-large;
    font-weight: bold;
    font-family:monospace;    
}

.txt_login{

    text-align: center;
    font-family:monospace;

    box-shadow: inset #abacaf 0 0 0 2px;
    border: 0;
    background: rgba(0, 0, 0, 0);
    appearance: none;
    position: relative;
    border-radius: 3px;
    padding: 9px 12px;
    line-height: 1.4;
    color: rgb(0, 0, 0);
    font-size: 16px;
    font-weight: 400;
    height: 40px;
    transition: all .2s ease;
    :hover{
        box-shadow: 0 0 0 0 #fff inset, #1de9b6 0 0 0 2px;
    }
    :focus{
        background: #fff;
        outline: 0;
        box-shadow: 0 0 0 0 #fff inset, #1de9b6 0 0 0 3px;
    }
}
::placeholder {
    color: #888686;
    opacity: 1;
    font-weight: bold;
    font-style: oblique;
    font-family:monospace;   
}
.btn_login {
    padding: 2px 10px 2px 10px;
    background-color: #060472; 
    color: #FFFFFF;
    font-weight: bold;
    font-size: large;
    border-radius: 5px;
    font-family:monospace;
    text-decoration: none;
    border-style:  none;
}

.btn_home{
    padding: 2px 10px 2px 10px;
    background-color: #089a28; 
    color: #FFFFFF;
    font-weight: bold;
    font-size: large;
    border-radius: 5px;
    font-family:monospace;
    text-decoration: none;
}

.btn_board {
    padding: 2px 10px 2px 10px;
    background-color: #934377; 
    border-style: none;
    color: #FFFFFF;
    font-weight: bold;
    font-size: large;
    border-radius: 5px;
    font-family:monospace;
    text-decoration: none;
}

.userword{
    color: #12103c;
    font-weight: bold;
    font-family:monospace;    
    font-size: xxx-large;
}

.files_status{
    font-weight: bold;
    font-size: x-large;
    font-family:monospace;
}

.files_list_up{
    padding: 10px 10px;
    background-color: #fcffa6; 
    color: #080000;
    font-size: x-large;
    border-radius: 10px;
    font-family:monospace;
    text-decoration: none;
}

.files_list_down{
    padding: 10px 10px;
    background-color: #ececec; 
    color: #080000;
    font-size: x-large;
    font-weight: bold;
    border-radius: 10px;
    font-family:monospace;
    text-decoration: none;
}

.msg_login{
    color: #060472; 
    font-size: large;
    font-weight: bold;
    font-family:monospace;    
    animation-duration: 3s; 
    animation-name: fader_msg;
}
@keyframes fader_msg {from {color: #ffffff;} to {color: #060472; } }


#file {
    border-style: solid;
    border-radius: 10px;
    font-family:monospace;
    background-color: #232323;
    border-color: #232323;
    color: #FFFFFF;
    font-size: small;
}
#submit {
    padding: 2px 10px 2px;
    background-color: #232323; 
    color: #FFFFFF;
    font-family:monospace;
    font-weight: bold;
    font-size: large;
    border-style: solid;
    border-radius: 10px;
    border-color: #232323;
    text-decoration: none;
    font-size: small;
}
#submit:hover {
  box-shadow: 0 12px 16px 0 rgba(0, 0, 0,0.24), 0 17px 50px 0 rgba(0, 0, 0,0.19);
}

"""
)


# ******************************************************************************************
HTML_TEMPLATES = dict(
# ******************************************************************************************

# ******************************************************************************************

login = """
<html>
    <head>
        <meta charset="UTF-8">
        <title> {{ config.title }} </title>
        <link rel="stylesheet" href="{{ url_for('static', filename='style.css') }}">  
    </head>
    <body>
    <!-- ---------------------------------------------------------->
    </br>
    <!-- ---------------------------------------------------------->

    <div align="center">
        <br>
        <div class="title">{{ config.title }}</div>
        <br>
        <br>
        <form action="{{ url_for('route_login') }}" method="post">
            <br>
            <div style="font-size: x-large;">{{ warn }}</div>
            <br>
            <div class="msg_login">{{ msg }}</div>
            <br>
            <input id="uid" name="uid" type="text" placeholder="... user-id ..." class="txt_login"/>
            <br>
            <br>
            <input id="passwd" name="passwd" type="password" placeholder="... password ..." class="txt_login"/>
            <br>
            <br>
            <input type="submit" class="btn_login" value="Login"> 
            <br>
            <br>
        </form>
    </div>

    <!-- ---------------------------------------------------------->
    <!-- ---------------------------------------------------------->
    </body>
</html>
""",
#******************************************************************************************
store = """
<html>
    <head>
        <meta charset="UTF-8">
        <title> {{ config.title }} | {{ session.uid }} </title>
        <link rel="stylesheet" href="{{ url_for('static', filename='style.css') }}">           
    </head>
    <body>
    <!-- ---------------------------------------------------------->
    </br>
    <!-- ---------------------------------------------------------->
    
    <div align="left" style="padding: 20px;">
        <a href="{{ url_for('route_logout') }}" class="btn_login">Logout</a>
        <br>
        <div class="userword">{{session.uid}}</div>
        <br>
        <!-- Breadcrumb for navigation -->
        <div class="files_status">
            {% if subpath %}
                <a href="{{ url_for('route_store') }}" class="btn_home">{{ config.rootname }}</a>{% for part in subpath.split('/') %}🔹<a href="{{ url_for('route_store', subpath='/'.join(subpath.split('/')[:loop.index])) }}" class="btn_board">{{ part }}</a>{% endfor %} <a href="{{ url_for('route_store', subpath=subpath, get='') }}" >⬇️</a> <a href="{{ url_for('route_store', subpath=subpath, del='') }}" >❌</a> 
            {% else %}
                <a href="{{ url_for('route_store') }}" class="btn_home">{{ config.rootname }}</a>
            {% endif %}
        </div>
        <!-- Directory Listing -->
        <br>
        <div class="files_list_up">
            {% for dir in dirs %}
            <a href="{{ url_for('route_store', subpath=subpath + '/' + dir) }}" class="btn_board">{{ dir }}</a>
            {% endfor %}
        </div>
        <hr>
        <form method='POST' enctype='multipart/form-data'>
            {{form.hidden_tag()}}
            {{form.file()}}
            {{form.submit()}}
        </form>
        <hr>
        <div class="files_list_down">
            <ol>
            {% for file in files %}
            <li>
            <a href="{{ url_for('route_store', subpath=subpath + '/' + file, get='') }}" >⬇️</a> 
            <a href="{{ url_for('route_store', subpath=subpath + '/' + file, del='') }}" >❌</a> 
            <a href="{{ url_for('route_store', subpath=subpath + '/' + file) }}" target="_blank">{{ file }}</a>
            </li>
            
            {% endfor %}
            </ol>
        </div>
        <br>
        <br>
    </div>

    <!-- ---------------------------------------------------------->
    </br>
    <!-- ---------------------------------------------------------->
    </body>
</html>
""",
)
# ******************************************************************************************

# ******************************************************************************************
# ******************************************************************************************


TEMPLATES_DIR, STATIC_DIR = os.path.join(PYDIR, "templates"), os.path.join(PYDIR, "static")
os.makedirs(TEMPLATES_DIR, exist_ok=True)
#sprint(f'↪ Creating html templates @ {TEMPLATES_DIR}')
for k,v in HTML_TEMPLATES.items():
    h = os.path.join(TEMPLATES_DIR, f"{k}.html")
    with open(h, 'w', encoding='utf-8') as f: f.write(v)
os.makedirs(STATIC_DIR, exist_ok=True)
#sprint(f'↪ Creating css templates @ {STATIC_DIR}')
for k,v in CSS_TEMPLATES.items():
    h = os.path.join(STATIC_DIR, f"{k}.css")
    with open(h, 'w', encoding='utf-8') as f: f.write(v)
sprint(f'↪ Created html/css templates @ {PYDIR}')


# ******************************************************************************************
#%%
# ------------------------------------------------------------------------------------------
# application setting and instance
# ------------------------------------------------------------------------------------------
app = Flask(__name__)
app.secret_key =          APP_SECRET_KEY
app.config['root'] =      ROOTDIR
app.config['rootname'] =  os.path.basename(ROOTDIR)
app.config['title'] =     parsed.title
app.config['case'] =      parsed.case
app.config['hide'] =      parsed.hide
# ------------------------------------------------------------------------------------------
class UploadFileForm(FlaskForm): # The upload form using FlaskForm
    file = MultipleFileField("File", validators=[InputRequired()])
    submit = SubmitField("Upload File")
# ------------------------------------------------------------------------------------------
@app.route('/', methods =['GET', 'POST'])
def route_login():
    LOGIN_NEED_TEXT =       '🔒'
    LOGIN_FAIL_TEXT =       '❌'     
    LOGIN_NEW_TEXT =        '🔥'
    LOGIN_CREATE_TEXT =     '🔑'    
    
    if request.method == 'POST' and 'uid' in request.form and 'passwd' in request.form:
        global db
        in_uid = f"{request.form['uid']}"
        in_passwd = f"{request.form['passwd']}"
        in_query = in_uid if not app.config['case'] else (in_uid.upper() if app.config['case']>0 else in_uid.lower())
        valid_query = AP.VALIDATE_UID(in_query)
        record = None if not valid_query else (db[in_query] if (in_query in db) else ...)
        if (record is None) or (record is ...):  warn, msg = LOGIN_FAIL_TEXT, f'[{in_uid}] Not a valid user'
        else: 
            access, uid, passwd = record
            if not passwd: # fist login
                if in_passwd: # new password provided
                    if AP.VALIDATE_PASS(in_passwd): # new password is valid
                        db[uid][-1]=in_passwd 
                        warn = LOGIN_CREATE_TEXT
                        msg = f'[{in_uid}] New password was created successfully'
                        dprint(f'● {in_uid} just joined via {request.remote_addr}')
                    else: # new password is invalid valid 
                        warn, msg = LOGIN_NEW_TEXT, f'[{in_uid}] New password is invalid - can use alpha-numeric, underscore and @-symbol'
                else: #new password not provided                
                    warn, msg = LOGIN_NEW_TEXT, f'[{in_uid}] New password required - can use alpha-numeric, underscore and @-symbol'
                                           
            else: # re login
                if in_passwd: # password provided 
                    if in_passwd==passwd:
                        session['has_login'] = True
                        session['uid'] = uid
                        session['access'] = access 
                        dprint(f'● {session["uid"]} has logged in via {request.remote_addr}') 
                        return redirect(url_for('route_store'))
                    else:  
                        warn, msg = LOGIN_FAIL_TEXT, f'[{in_uid}] Password mismatch'             
                else: # password not provided
                    warn, msg = LOGIN_FAIL_TEXT, f'[{in_uid}] Password not provided'
    else:
        if session.get('has_login', False):  return redirect(url_for('route_store'))
        warn, msg = LOGIN_NEED_TEXT, f'Login to continue'
        
        
    return render_template('login.html', msg = msg,  warn = warn)

@app.route('/logout')
def route_logout():
    r""" logout a user and redirect to login page """
    if not session.get('has_login', False):  return redirect(url_for('route_login'))
    if not session.get('uid', False): return redirect(url_for('route_login'))
    if session['has_login']:  dprint(f'● {session["uid"]} has logged out via {request.remote_addr}') 
    else: dprint(f'✗ {session["uid"]}was removed due to invalid uid via {request.remote_addr}') 
    session.clear()
    return redirect(url_for('route_login'))
# ------------------------------------------------------------------------------------------
TOKEN_DOWNLOAD = 'get'
TOKEN_DELETE = 'del'
import shutil
# ------------------------------------------------------------------------------------------
@app.route('/store', methods =['GET', 'POST'])
@app.route('/store/', methods =['GET', 'POST'])
@app.route('/store/<path:subpath>', methods =['GET', 'POST'])
def route_store(subpath=""):
    if not session.get('has_login', False): return redirect(url_for('route_login'))
    
    form = UploadFileForm()
    abs_path = os.path.join(app.config['root'], subpath)
    if not os.path.exists(abs_path): return abort(404)

    if os.path.isdir(abs_path):
        
        if form.validate_on_submit() and ('U' in session['access']):
            dprint(f"● {session['uid']} is trying to upload {len(form.file.data)} items via {request.remote_addr}")
        
            result = []
            n_success = 0
            #---------------------------------------------------------------------------------
            for file in form.file.data:
                isvalid, sf = FP(secure_filename(file.filename))
            #---------------------------------------------------------------------------------
                
                if not isvalid:
                    why_failed =  f"✗ Extension is invalid [{sf}] "
                    result.append((0, why_failed))
                    continue

                file_name = os.path.join(abs_path, sf)
                if os.path.exists(file_name):
                    if ('R' in session['access']):
                        try:
                            file.save(file_name) 
                            why_failed = f"✓ Uploaded new file [{sf}] "
                            result.append((1, why_failed))
                            n_success+=1
                        except PermissionError:
                            why_failed = f"✗ Permission denied [{sf}] "
                            result.append((0, why_failed))


            #---------------------------------------------------------------------------------
                
            result_show = ''.join([f'\t{r[-1]}\n' for r in result])
            result_show = result_show[:-1]
            dprint(f'✓ {session["uid"]} just uploaded {n_success} file(s)\n{result_show}') 
        
        
        items = os.listdir(abs_path)
        if app.config['hide']: items = [item for item in items if not item.startswith('.')]
        dirs = [item for item in items if os.path.isdir(os.path.join(abs_path, item))]
        files = [item for item in items if os.path.isfile(os.path.join(abs_path, item))]
        if TOKEN_DOWNLOAD in request.args:
            os.chdir(app.config['root'])
            zp = os.path.relpath(abs_path, app.config['root'])
            Z = ZipFolders(zp) # zipped, zip_path
            os.chdir(CWDIR)
            #print(zipped, zip_path)
            dprint(f'✓ {session["uid"]} zipped {zp} file(s) -> {Z}') 
            return redirect(url_for('route_store', subpath=zp+'.zip'))
        elif TOKEN_DELETE in request.args:
            try:
                shutil.rmtree(abs_path, False)
                dprint(f'✓ {session["uid"]} deleted {abs_path}') 
                return redirect(url_for('route_store', subpath=os.path.dirname(subpath)))
            except:
                dprint(f'✓ {session["uid"]} could not delete {abs_path}') 
                return redirect(url_for('route_store', subpath=subpath))
        else: pass
        return render_template('store.html', dirs=dirs, files=files, subpath=subpath, form=form)
    elif os.path.isfile(abs_path): 
        if 'D' not in session['access']:  return abort(404)
        else:  
            #print (request.args)
            if request.args:
                if TOKEN_DOWNLOAD in request.args:
                    dprint(f'● {session["uid"]} downloaded {abs_path} via {request.remote_addr}')
                    return send_file(abs_path, as_attachment=True)
                elif TOKEN_DELETE in request.args:
                    if 'X' not in session['access']:  return redirect(url_for('route_store', subpath=os.path.dirname(subpath)))
                    try:
                        os.remove(abs_path)
                        dprint(f'● {session["uid"]} deleted {abs_path} via {request.remote_addr}')
                    except:
                        dprint(f'● {session["uid"]} could not delete {abs_path} via {request.remote_addr}')
                    return redirect(url_for('route_store', subpath=os.path.dirname(subpath)))
            else : return send_file(abs_path, as_attachment=False)
    else: return abort(404)




# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#<-------------------DO NOT WRITE ANY NEW CODE AFTER THIS
endpoint = f'{parsed.host}:{parsed.port}' if parsed.host!='0.0.0.0' else f'localhost:{parsed.port}'
sprint(f'◉ http://{endpoint}')
start_time = datetime.datetime.now()
sprint('◉ start server @ [{}]'.format(start_time))
serve(app, # https://docs.pylonsproject.org/projects/waitress/en/stable/runner.html
    host = parsed.host,          
    port = parsed.port,          
    url_scheme = 'http',     
    threads = parsed.threads,    
    connection_limit = parsed.mcon,
    max_request_body_size = FP.MAX_UPLOAD_SIZE,
)
#<-------------------DO NOT WRITE ANY CODE AFTER THIS
end_time = datetime.datetime.now()
sprint('◉ stop server @ [{}]'.format(end_time))
try: 
    db > LOGINFILE
    sprint('↷ persisted login-db')
except: sprint('↷ could not persist login-db')

sprint(f'↪ Cleaning up html/css templates...')
try:
    for k,v in HTML_TEMPLATES.items():
        h = os.path.join(TEMPLATES_DIR, f"{k}.html")
        if  os.path.isfile(h) : os.remove(h)
    #sprint(f'↪ Removing css templates @ {STATIC_DIR}')
    for k,v in CSS_TEMPLATES.items():
        h = os.path.join(STATIC_DIR, f"{k}.css")
        if os.path.isfile(h): os.remove(h)
    os.removedirs(TEMPLATES_DIR)
    os.removedirs(STATIC_DIR)
    sprint(f'↪ Removed html/css templates @ {PYDIR}')
except:
    sprint(f'↪ Could not remove html/css templates @ {PYDIR}')
sprint('◉ server up-time was [{}]'.format(end_time - start_time))
sprint(f'...Finished!')
# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


__doc__=r"""pix execute"""

#-----------------------------------------------------------------------------------------
from sys import exit
if __name__!='__main__': exit(f'[!] Can not import {__name__}:{__file__}')
#-----------------------------------------------------------------------------------------

import os, json, argparse
from . import Actions

avaliable_actions = [k for k in Actions.__dict__ if not k.startswith('__')]
#-----------------------------------------------------------------------------------------

def _read_fl_from_text(path):
    with open(path, 'r') as f: l = [ os.path.abspath(f'{s}') for s in f.read().split('\n') if s ]
    return l
def _read_fl_from_json(path):
    with open(path, 'rb') as f: l = json.load(f)
    return l
def _read_fl_from_linux(F): # parses --files="%F"
    Fl = [fi.strip() for fi in F.split("'/")]
    Fr = [os.path.abspath(f'/{fl[:-1]}'.replace("'\\''","'")) for fl in Fl if fl] 
    return Fr
def _read_fl(parsed_put, parsed_io, check=False):
    if parsed_put:
        _put = os.path.abspath(parsed_put)
        if not parsed_io:           _puts = [_put] 
        elif parsed_io == 't':   _puts =_read_fl_from_text(_put)
        elif parsed_io == 'j':   _puts =_read_fl_from_json(_put)
        elif parsed_io == 'l':   _puts =_read_fl_from_linux(_put)
        else:                       _puts = [] 
    else:                           _puts = []
    if check: _puts = [p for p in _puts if os.path.isfile(p)]
    return _puts


#-----------------------------------------------------------------------------------------

# actions = new, crop, extend, flip, rotate, convert
parser = argparse.ArgumentParser()
parser.add_argument('--action', type=str, default='',   help=f"(str) one of the static-methods inside the Actions class, can be - {avaliable_actions}")
parser.add_argument('--args',   type=str, default='',   help="(str) csv args accepted by the specified action - each action takes different args")
parser.add_argument('--input',  type=str,   default='', help='(str) input  image-file or a text/json-file containing multiple input image-file names') 
parser.add_argument('--output',  type=str,  default='', help='(str) output image-file or a text/json-file containing multiple output image-file names') 
parser.add_argument('--files',  type=str,   default='', help='(str) multiple input image-file names - for custom action -- works only with --io=linux') 
parser.add_argument('--io',      type=str,  default='', help="(str) can be 'text' or 'json' or 'linux' - keep blank to io as 'image' - used if providing input/output file-names in a text/json file")
parser.add_argument('--verbose', type=int,  default=0,  help="(int) verbose level - 0 or 1")

parsed = parser.parse_args()

# ---------------------------------------------------------------------------------
_verbose = int(parsed.verbose)
_action = f'{parsed.action}'
if not _action: exit(f'[!] Action not provided')
if not hasattr(Actions, _action): exit(f'[!] Action [{_action}] not found')
_action = getattr(Actions, _action)
# ---------------------------------------------------------------------------------
_args = f'{parsed.args}'.split(',')
# ---------------------------------------------------------------------------------
_io = f'{parsed.io}'.lower()[0]
if _io == 'l':
    _inputs =   _read_fl(f'{parsed.files}',  _io, check=True) # assume existing files are passed
    _outputs =  _inputs
else:
    _inputs =   _read_fl(f'{parsed.input}',  _io, check=True) # keep only existing files
    _outputs =  _read_fl(f'{parsed.output}', _io, check=False)
    if not _outputs: _outputs = _inputs # if outputs are not provided, overwrite inputs
    if _inputs: assert len(_inputs) == len(_outputs), f'Mismatch inputs and outputs' # if inputs were provided, outputs must match them

# ---------------------------------------------------------------------------------
_action(_inputs, _outputs, _args, _verbose)
# ---------------------------------------------------------------------------------




__doc__=f""" 
-------------------------------------------------------------
topics - Flask-based web app for sharing files 
-------------------------------------------------------------

Install the required packages
    
    python -m pip install Flask Flask-WTF waitress nbconvert

Note: the `nbconvert` is optional - without it, the 'board' functionality will be missing

To strat the server use,

    python -m known.topics --dir=path/to/workspace

To stop the server, press control-c or send keyboard interupt signal

Note: default `configs.py` file will be created under workspace directory
-> the dict named `current` will be choosen as the config - it should be defined at the end of the file
-> a config name `default` will be created - it is used as a fall-back config

Edit the configs file which should be located at `path/to/workspace/configs.py`
Multiple configs can be defined in that single file, only the `current` variable will be used

-------------------------------------------------------------
Note:
special string "::::" is used for replacing javascript on `repass` - uid and url should not contain this
special username 'None' is not allowed however words like 'none' will work
rename argument means (0 = not allowed) (1 = only rename) (2 = rename and remoji)
-------------------------------------------------------------
"""






import os
from zipfile import ZipFile

def ZipFiles(files:list, zip_path:str=None, **kwargs):
    r""" zips all (only files) in the list of file paths and saves at 'zip_path' """
    if isinstance(files, str): files= [f'{files}']
    if not files: return # no files provided to zip
    if not zip_path : zip_path = f'{files[0]}.zip' # if zip path not provided take it to be the first file
    if not zip_path.lower().endswith('.zip'): zip_path = f'{zip_path}.zip'  # append .zip to the end of path

    zipped = 0
    with ZipFile(zip_path, 'w', **kwargs) as zip_object:
        for path in files:
            if not os.path.isfile(path): continue
            zip_object.write(f'{path}')
            zipped+=1
    return zipped, zip_path

def ZipFolders(folders, zip_path:str=None, **kwargs):  
    r""" zip multiple folders into a single zip file 
    to zip a single folder with the same zip name - provide folder as a string and keep zip_path as none
    """    
    if isinstance(folders, str): folders= [f'{folders}']
    if not folders: return None, None# no folders provided to zip
    if not zip_path : zip_path = f'{folders[0]}.zip' # if zip path not provided take it to be the first folder
    if not zip_path.lower().endswith('.zip'): zip_path = f'{zip_path}.zip'  # append .zip to the end of path
    all_files = []
    for folder in folders:
        for root, directories, files in os.walk(folder): all_files.extend([os.path.join(root, filename) for filename in files])
    return ZipFiles(all_files, f'{zip_path}', **kwargs)

class Table:

    @staticmethod
    def CreateData(*columns):
        data = {None:[f'{col}' for col in columns]} # this is to make sure that col names are always on top
        return data

    @staticmethod
    def Create(columns:tuple, primary_key:str, cell_delimiter=',', record_delimiter='\n'):
        # should be called on a new object after init\
        table = __class__()
        table.data = __class__.CreateData(*columns)
        table.pk = primary_key
        table.pkat = table.data[None].index(table.pk)
        table.cell_delimiter, table.record_delimiter = cell_delimiter, record_delimiter
        return table


    @staticmethod
    def ImportData(path, key_at, cell_delimiter, record_delimiter): 
        with open(path, 'r', encoding='utf-8') as f: 
            s = f.read()
            lines = s.split(record_delimiter)
            cols = lines[0].split(cell_delimiter) #<--- only if None:cols was added as a first entry (using Create method)
            data = {None:cols}
            if isinstance(key_at, str): key_at = cols.index(key_at)
            assert key_at>=0,f'Invlaid key {key_at}'
            for line in lines[1:]:
                if line:
                    cells = line.split(cell_delimiter)
                    data[f'{cells[key_at]}'] = cells
        return data
    
    @staticmethod
    def Import(path, key_at, cell_delimiter=',', record_delimiter='\n'): 
        table = __class__()
        table.data = __class__.ImportData(path, key_at, cell_delimiter, record_delimiter)
        if isinstance(key_at, str): key_at = table[None].index(key_at)
        table.pk = table.data[None][key_at]
        table.pkat = key_at
        table.cell_delimiter, table.record_delimiter = cell_delimiter, record_delimiter
        return table


    @staticmethod
    def ExportData(data, path, cell_delimiter, record_delimiter): 
        with open(path, 'w', encoding='utf-8') as f: 
            for v in data.values(): f.write(cell_delimiter.join(v)+record_delimiter)

    @staticmethod
    def Export(table, path): 
        __class__.ExportData(table.data, path, table.cell_delimiter, table.record_delimiter)

    # get row as dict
    def __call__(self, key): return {k:v for k,v in zip(self[None], self[key])}

    # get row as it is (list)
    def __getitem__(self, key): return self.data[key]

    # set row based on if its a dict or a list (note: key is irrelavant here)
    def __setitem__(self, key, row):
        assert len(row) == len(self[None]), f'Rows are expected to have length {len(self[None])} but got {len(row)}'
        if isinstance(row, dict):
            key = row[self.pk]
            if key is not None: self.data[f'{key}'] = [row[r] for r in self[None]]
        else: 
            key = row[self.pkat]
            if key is not None: self.data[f'{key}'] = list(row)

    # del row based on key
    def __delitem__(self, key):
        if key is not None: del self.data[key]

    def __contains__(self, key): return key in self.data

    # quick export > file
    def __gt__(self, other):__class__.ExportData(self.data, f'{other}', self.cell_delimiter, self.record_delimiter)

    # quick import < file
    def __lt__(self, other): self.data = __class__.ImportData(f'{other}', self.pkat, self.cell_delimiter, self.record_delimiter)

    # total number of rows
    def __len__(self): return len(self.data)-1






__doc__="""
HTTP API Request-Response model.
This module is designed to enable communication between a remote python process and a central server using HTTP API.
"""

#-----------------------------------------------------------------------------------------
from sys import exit
if __name__!='__main__': exit(f'[!] Can not import {__name__}:{__file__}')
#-----------------------------------------------------------------------------------------



#-----------------------------------------------------------------------------------------
# imports 
#-----------------------------------------------------------------------------------------
import os, argparse, datetime, importlib, sys
from . import HeaderType, RequestContentType, ResponseContentType, StoreType, DefaultArgs

#PYDIR = os.path.dirname(__file__) # script directory of __main__.py
try:

    from flask import Flask, request, send_file
    from waitress import serve
    from http import HTTPStatus
    from shutil import rmtree
except: exit(f'[!] Required packages missing - pip install Flask waitress')
#-----------------------------------------------------------------------------------------



# ==============================================================================================================
# Common Functions 
# NOTE: common functions are repeated in all modular servers so that they can act as stand alone
# ==============================================================================================================

class Fake:
    def __len__(self): return len(self.__dict__)
    def __init__(self, **kwargs):
        for k,v in kwargs.items():setattr(self, k, v)
    #def _get_kwargs(self): return self.__dict__.items()

class HRsizes: # human readable size like  000.000?B
    mapper = dict(KB=2**10, MB=2**20, GB=2**30, TB=2**40)
    def tobytes(size): return int(float(size[:-2])*__class__.mapper.get(size[-2:].upper(), 0))

class EveryThing: # use as a set that contains everything (use the 'in' keyword)
    def __contains__(self, x): return True

# ==============================================================================================================
DEFAULT_CONFIG_CODE = """
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# known.API configuration file                                                                                   ||
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~




# =================================================================================================================
# the running configuration - must be a dict named as variable 'config'                                          ||
# =================================================================================================================
config = {
    'maxH'        : "0.25GB",               # (str) maximum http header size
    'maxB'        : "1.0GB",                # (str) maximum http body size
    'limit'       : 5,                      # (int) maximum connection limit to the server
    'port'        : "8888",                 # (str) server port
    'host'        : "0.0.0.0",              # (str) server address (keep 0.0.0.0 to run on all IPs)
    'allow'       : "",                     # (str) a comma-seperated list of host IP address that are allowed to POST (keep blank to allow all)
    'threads'     : 10,                     # (int) no of threads on the server
    'access'      : 'access',               # (str) the access dict variable name (as string) defined in this file 
}
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
# Note: we can define multiple `access` dicts and then choose in running config 
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# =================================================================================================================




# =================================================================================================================
# the access configuration - must be a dict specified in config['access']                                        ||
# =================================================================================================================
access = { 
#   username    :   (   access-string,      storage_path,       handle_function      ),
    'root'      :   (   'AGPD',             '!',                'handle'             ),  
}
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
# Note: 
# by default 
#   a `root` user is created 
#   storage is set to the workdir excluding top level files
# `access` is a dict of { user_name : (access_string, storage_path, handle_function) }
# -> if access dict is blank, no users will be allowed
# -> `user_name` is a string specifying the user
#     --> `access_string` is a string specifying the access permissions of the user
#     --> `storage_path` is a string specifying absolute path which acts as a base-dir for that user 
#             ~ keep blank to use directory specified in the `--dir` argument
#             ~ prefix with '!' to exclude top-level files
#             ~ set as None or ... or anything not str to disable storage
#     --> `handle_function` is a callable like `lambda request_content, request_type: (response_content, response_tag)`
#             ~ handle_function must be define in this file as well
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
# access_string can be made up of the following:
#   (A)- can access API vis POST request (at root url)
#   (G)- can download files and can list folder contents (GET)
#   (P)- can write/overwrite files and create folder (PUT, POST)
#   (D)- can delete files and folder (DELETE)
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
# =================================================================================================================





# =================================================================================================================
# handle functions                                                                                               ||
# =================================================================================================================
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
# Note: this function handles an incoming request from client - these are POST requests directly to the root url 
# use arguments `request_type` and `request_content` to access the client request data and process the data ...
# return a valid response containing a 2-tuple (response_content, response_tag)
# `response_content` should be one of (str, bytes, dict, list, tuple )
# `response_tag` is just a string
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
# =================================================================================================================

def handle(request_content:object, request_type:str):
    if request_content is None : ... # request_content is None
    elif request_type == 'MESG': ... # request_content is string   (   request_content = request.get_data().decode('utf-8') )
    elif request_type == 'BYTE': ... # request_content is bytes    (   request_content = request.get_data()                 )
    elif request_type == 'FORM': ... # request_content is 2-tuple  (   request_content = (request.form, request.files)      )
    elif request_type == 'JSON': ... # request_content is json     (   request_content = request.get_json()                 )
    else                       : ... # request_content is unknown
    response_content, response_tag = "default_response", "default_tag"
    return response_content, response_tag

# =================================================================================================================

"""


#-----------------------------------------------------------------------------------------
# Parse arguments 
#-----------------------------------------------------------------------------------------
parser = argparse.ArgumentParser()
for k,v in DefaultArgs.items(): parser.add_argument(f'--{k}', **v)
parsed = parser.parse_args()
#-----------------------------------------------------------------------------------------

WORKDIR = f'{parsed.dir}'                               # define working dir - contains all bases
if not WORKDIR: WORKDIR = os.getcwd()    # if still not specified, set as default
print(f'↪ Workspace directory is {WORKDIR}')
try: os.makedirs(WORKDIR, exist_ok=True)
except: exit(f'[!] Workspace directory was not found and could not be created')
sys.path.append(WORKDIR)  # otherwise append to sys.path

#-----------------------------------------------------------------------------------------
# ==> read configurations
#-----------------------------------------------------------------------------------------
CONFIG = 'config'
CONFIG_MODULE = parsed.config if parsed.config else 'api' # the name of configs module
CONFIGS_FILE = f'{CONFIG_MODULE}.py' # the name of configs file


# check if 'configs.py` exsists or not`
CONFIGS_FILE_PATH = os.path.join(WORKDIR, CONFIGS_FILE) # should exsist under workdir
first_exit=False
if not os.path.isfile(CONFIGS_FILE_PATH):
    print(f'↪ Creating default config "{CONFIGS_FILE}" from ...')
    with open(CONFIGS_FILE_PATH, 'w', encoding='utf-8') as f: f.write(DEFAULT_CONFIG_CODE)
    first_exit=True
try: 
    c_module = importlib.import_module(f'{CONFIG_MODULE}')
    print(f'↪ Imported config-module "{CONFIG_MODULE}" from {c_module.__file__}')
except: exit(f'[!] Could import configs module "{CONFIG_MODULE}" at "{CONFIGS_FILE_PATH[:-3]}"')
try:
    print(f'↪ Reading config from {CONFIG_MODULE}.{CONFIG}')
    config_dict = getattr(c_module, CONFIG)
    #print(f'  ↦ type:{type(config_dict)}')
except:
    exit(f'[!] Could not read config from {CONFIG_MODULE}.{CONFIG}')

if not isinstance(config_dict, dict): 
    try: config_dict=config_dict()
    except: pass
if not isinstance(config_dict, dict): raise exit(f'Expecting a dict object for config')

try: 
    print(f'↪ Building config from {CONFIG_MODULE}.{CONFIG}')
    #for k,v in config_dict.items(): print(f'  ↦ {k}\t\t{v}')
    args = Fake(**config_dict)
except: exit(f'[!] Could not read config')
if not len(args): exit(f'[!] Empty or Invalid config provided')


ACCESS = args.access
try:
    print(f'↪ Getting access dict from {CONFIG_MODULE}.{ACCESS}')
    user_access = getattr(c_module, ACCESS)
    #print(f'  ↦ type:{type(user_access)}')
except:
    exit(f'[!] Could not get access dict from {CONFIG_MODULE}.{ACCESS}')

user_handles = {}
try:
    for k,(_,_,HANDLE) in user_access.items():
        print(f'↪ Getting handle from {CONFIG_MODULE}.{HANDLE}')
        user_handles[k] = getattr(c_module, HANDLE)
        #print(f'  ↦ type:{type(user_handles[k])}')
except:
    exit(f'[!] Could not get handle from {CONFIG_MODULE}.{HANDLE}')


if first_exit:
    print(f'↪ Configuration was built! Press [enter] to start server now ↦↦↦↦')
    if input(): exit('Server was not started!')
# ------------------------------------------------------------------------------------------
# application setting and instance
# ------------------------------------------------------------------------------------------
app = Flask(__name__)
if args.allow:
    allowed = set(args.allow.split(','))
    if '' in allowed: allowed.remove('')
else: allowed = EveryThing()


def GET_ACCESS_CONFIG(access_dict):
    ac_access, ac_storage, ac_xtop = {}, {}, {}
    for user_name,(access_string,storage_path,_) in access_dict.items():
        ac_access[user_name] = access_string.upper()

        if not isinstance (storage_path, str):  ac_xtop[user_name], ac_storage[user_name] =    True,  None # no storage
        else:
            if storage_path.startswith('!'):      ac_xtop[user_name], ac_storage[user_name] =    True,  storage_path[1:]
            else:                                 ac_xtop[user_name], ac_storage[user_name] =    False, storage_path       
            if ac_storage[user_name]=='':         ac_storage[user_name] = os.path.abspath(WORKDIR) # workdir
            else:                                 ac_storage[user_name] = os.path.abspath(ac_storage[user_name]) # abspath
    return ac_access, ac_storage, ac_xtop
        



app.config['allow'] = allowed
app.config['access'], app.config['storage'], app.config['xtop']  =  GET_ACCESS_CONFIG(user_access)
#-----------------------------------------------------------------------------------------

def sprint(uid, method, path): print('↦ [{}]\t{}\t({})\t{}'.format(datetime.datetime.now(), method, uid, path))

#-----------------------------------------------------------------------------------------
# NOTE on return type
# ... type must be a string, dict, list, 
# ... type can be this but ignored: tuple with headers or status, Response instance, or WSGI callable
#-----------------------------------------------------------------------------------------
@app.route('/', methods =['GET', 'POST'])
def home():
    global user_handles
    
    if request.method == 'POST':
        request_from = request.environ['REMOTE_HOST'] 
        if request_from in app.config['allow']:

            # The clients making post request will have to provide these two headers
            # headers are only read for post requests from allowed users
            xtag, xtype = request.headers.get(HeaderType.XTAG), request.headers.get(HeaderType.XTYPE)
            if xtag not in app.config['access']: xcontent = None
            else:
                if 'A' not in app.config['access'][xtag]: xcontent = None
                elif xtype is None:             xcontent = None
                #------------------------------------------------------------------------------- Read from the reuest made by client
                elif xtype==RequestContentType.MESG: xcontent = request.get_data().decode('utf-8')
                elif xtype==RequestContentType.BYTE: xcontent = request.get_data()
                elif xtype==RequestContentType.FORM: xcontent = request.form, request.files
                elif xtype==RequestContentType.JSON: xcontent = request.get_json()
                #-------------------------------------------------------------------------------
                else:                         xcontent = None               
            
            if xcontent is not None:
                return_object, return_tag = user_handles[xtag](xcontent, xtype)
                return_type = ResponseContentType.ALL.get(type(return_object), None)
                if return_type is not None: 
                    return_code = HTTPStatus.OK
                    return_headers = {HeaderType.XTAG :f'{return_tag}', HeaderType.XTYPE:return_type} #<-- headers are only sent when content and types are valid
                    sprint(xtag, 'API', xtype)
                else:   return_object, return_code, return_headers = f"[!] Invalid response from handler [{type(return_object)}::{return_type}:{return_tag}]", HTTPStatus.NOT_FOUND, {}
            else:       return_object, return_code, return_headers = f'[!] Type "{xtype}" is not a valid content type', HTTPStatus.NOT_ACCEPTABLE, {}
        else:           return_object, return_code, return_headers = f"[!] You are not allowed to POST", HTTPStatus.NOT_ACCEPTABLE, {}
    elif request.method == 'GET':     
        return_object = f'<pre>[Known.api]@{__file__}\n[Workdir]@{WORKDIR}\n[Config]@{CONFIGS_FILE_PATH}</pre>'
        #for k,v in args._get_kwargs(): return_object+=f'\n\t{k}\t{v}\n'
        return_code, return_headers = HTTPStatus.OK, {}
    else: return_object, return_code, return_headers = f"[!] Invalid Request Type {request.method}", HTTPStatus.BAD_REQUEST, {}
    
    return return_object, return_code, return_headers


# Storage urls for file-storage api
# tag specifies a name of file, type specifies if its a overall-view, a directory listing or a file

@app.route('/store', methods =['GET'])
def storageview(): # an overview of all storage paths and the files in them
    uid = request.headers.get(HeaderType.XTAG)
    if uid not in app.config['access']:  return f'Invalid client uid [{uid}]', HTTPStatus.NOT_ACCEPTABLE, {}
    else:
        if 'G' not in app.config['access'][uid]:  return_object, return_code, return_headers = {}, HTTPStatus.NOT_ACCEPTABLE, {}
        else:
            basedir = app.config['storage'][uid]
            if basedir is None: return_object, return_code, return_headers = {}, HTTPStatus.NOT_FOUND, {}
            else: return_object, return_code, return_headers =  {os.path.relpath(root, basedir) : files for root, directories, files in os.walk(basedir)}, HTTPStatus.OK, {HeaderType.XTAG: f'{basedir}', HeaderType.XTYPE: StoreType.HOME}
        return return_object, return_code, return_headers

@app.route('/store/', methods =['GET'])
def storageroot(): # root dir
    uid = request.headers.get(HeaderType.XTAG)
    if uid not in app.config['access']:  return f'Invalid client uid [{uid}]', HTTPStatus.NOT_ACCEPTABLE, {}
    else:
        if 'G' not in app.config['access'][uid]:  return_object, return_code, return_headers = {}, HTTPStatus.NOT_ACCEPTABLE, {}
        else:
            basedir = app.config['storage'][uid]
            print(f'{basedir=}')
            if basedir is None: return_object, return_code, return_headers = {}, HTTPStatus.NOT_FOUND, {}
            else:
                rw, dw, fw = next(iter(os.walk(basedir)))
                fs = [os.path.getsize(os.path.join(rw, fwi)) for fwi in fw]
                rel_path = os.path.relpath(rw, basedir)
                return_object = dict(base=os.path.relpath(rw, basedir), folders=dw, files={k:round(v/1024,2) for k,v in zip(fw,fs)}) # size in KB
                return_code = HTTPStatus.OK
                return_headers = {HeaderType.XTAG: f'{rel_path}', HeaderType.XTYPE: StoreType.DIR}
    return return_object, return_code, return_headers

@app.route('/store/<path:req_path>', methods =['GET', 'POST', 'PUT', 'DELETE'])
def storage(req_path): # creates a FileNotFoundError

    uid = request.headers.get(HeaderType.XTAG)
    if uid not in app.config['access']:  return f'Invalid client uid [{uid}]', HTTPStatus.NOT_ACCEPTABLE, {}
    basedir = app.config['storage'][uid]
    if basedir is None: return_object, return_code, return_headers = f'No Storage is specified for Client uid [{uid}]', HTTPStatus.NOT_ACCEPTABLE, {}
    else:
        access_string = app.config['access'][uid]
        abs_path = os.path.join(basedir, req_path) # Joining the base and the requested path
        rel_path = os.path.relpath(abs_path, basedir)
        if request.method=='GET': # trying to download that file or view a directory
            if 'G' not in access_string: return_object, return_code, return_headers = f'GET Access denined to client uid [{uid}]', HTTPStatus.NOT_ACCEPTABLE, {}
            else:
                if os.path.exists(abs_path):
                    if os.path.isdir(abs_path):     
                        rw, dw, fw = next(iter(os.walk(abs_path)))
                        fs = [os.path.getsize(os.path.join(rw, fwi)) for fwi in fw]
                        return_object = dict(base=rel_path, folders=dw, files={k:round(v/1024,2) for k,v in zip(fw,fs)}) # size in KB
                        return_code = HTTPStatus.OK
                        return_headers = {HeaderType.XTAG: f'{rel_path}', HeaderType.XTYPE: StoreType.DIR}
                    else: 
                        # make sure to exclude top level files
                        if app.config['xtop'][uid]:
                            if (basedir==os.path.dirname(abs_path)): 
                                return f'Top-Level Path is excluded: {abs_path} - cannot read', HTTPStatus.NOT_ACCEPTABLE, {} #<-----RETURNING HERE

                        resx = send_file(abs_path) 
                        resx.headers[HeaderType.XTAG] = os.path.basename(abs_path) # 'save_as'
                        resx.headers[HeaderType.XTYPE] = StoreType.FILE
                        sprint(uid, 'STORE:DN', abs_path) # downloading a file 
                        return resx #<-----RETURNING HERE
                    
                else: return_object, return_code, return_headers = f'Path not found: {abs_path}', HTTPStatus.NOT_FOUND, {}


        elif request.method=='POST': # trying to create new file or replace existing file
            if 'P' not in access_string: return_object, return_code, return_headers = f'POST Access denined to client uid [{uid}]', HTTPStatus.NOT_ACCEPTABLE, {}
            else:
                if os.path.isdir(abs_path):
                    return_object, return_code, return_headers = f'Cannot create file # {abs_path} - folder already exists', HTTPStatus.NOT_ACCEPTABLE, {}
                else:
                    if app.config['xtop'][uid]:
                        if (basedir==os.path.dirname(abs_path)): 
                            return f'Top-Level Path is excluded: {abs_path} - cannot write', HTTPStatus.NOT_ACCEPTABLE, {} #<-----RETURNING HERE
                    try: # overwrite
                        with open(abs_path, 'wb') as f: f.write(request.get_data())
                        return_object, return_code, return_headers =  f"File created @ {abs_path}", HTTPStatus.OK, {HeaderType.XTAG: f'{rel_path}', HeaderType.XTYPE: StoreType.MSG}
                        sprint(uid, 'STORE:UP', abs_path) # uploading a file
                    except: return_object, return_code, return_headers =   f"Cannot create file @ {abs_path}", HTTPStatus.NOT_ACCEPTABLE, {}


        elif request.method=='PUT': # trying to create new directory
            if 'P' not in access_string: return_object, return_code, return_headers = f'PUT Access denined to client uid [{uid}]', HTTPStatus.NOT_ACCEPTABLE, {}
            else:
                if os.path.isfile(abs_path):
                    return_object, return_code, return_headers = f'Cannot create folder at {abs_path} - file already exists', HTTPStatus.NOT_ACCEPTABLE, {}
                else:
                    os.makedirs(abs_path, exist_ok=True)
                    return_object, return_code, return_headers =  f"Folder created @ {abs_path}", HTTPStatus.OK, {HeaderType.XTAG: f'{rel_path}', HeaderType.XTYPE: StoreType.MSG}
                    sprint(uid, 'STORE:MK', abs_path) # creating a folder


        elif request.method=='DELETE': # trying to delete a file or folder
            if 'D' not in access_string: return_object, return_code, return_headers = f'DELETE Access denined to client uid [{uid}]', HTTPStatus.NOT_ACCEPTABLE, {}
            else:
                if os.path.isfile(abs_path):
                    if app.config['xtop'][uid]:
                        if (basedir==os.path.dirname(abs_path)): 
                            return f'Top-Level Path is excluded: {abs_path} - cannot delete', HTTPStatus.NOT_ACCEPTABLE, {} #<-----RETURNING HERE
                    try: 
                        os.remove(abs_path)
                        return_object, return_code, return_headers =     f"File deleted @ {abs_path}", HTTPStatus.OK, {HeaderType.XTAG: f'{rel_path}', HeaderType.XTYPE: StoreType.MSG}
                        sprint(uid, 'STORE:DF', abs_path) # removing a file
                    except: return_object, return_code, return_headers = f"Cannot delete file @ {abs_path}", HTTPStatus.NOT_ACCEPTABLE, {}
                elif os.path.isdir(abs_path):
                    rok = True
                    rhead = request.headers.get(HeaderType.XTYPE)
                    if rhead is None: rhead = 0 # default behaviour is to not go for recursive delete
                    if int(rhead):
                        try: 
                            rmtree(abs_path)
                            sprint(uid, 'STORE:DR', abs_path) # removing a directory (recursive)
                        except: rok=False
                    else:
                        try: 
                            os.rmdir(abs_path)
                            sprint(uid, 'STORE:DE', abs_path) # removing a directory (empty)
                        except: rok=False
                    if rok: 
                        return_object, return_code, return_headers =   f"Folder deleted @ {abs_path}", HTTPStatus.OK, {HeaderType.XTAG: f'{rel_path}', HeaderType.XTYPE: StoreType.MSG}
                        
                    else:   return_object, return_code, return_headers =   f'Cannot delete folder at {abs_path}', HTTPStatus.NOT_ACCEPTABLE, {}
                else: return_object, return_code, return_headers =         f'Cannot delete at {abs_path} - not a file or folder', HTTPStatus.NOT_ACCEPTABLE, {}

        else: return_object, return_code, return_headers =  f"[!] Invalid Request Type {request.method}", HTTPStatus.BAD_REQUEST, {}

    return return_object, return_code, return_headers




#%% @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
start_time = datetime.datetime.now()
print('◉ start server @ [{}]'.format(start_time))
# print(f'{app.config["access"]}')
# print(f'{app.config["storage"]}')
# print(f'{app.config["xtop"]}')
serve(app, # https://docs.pylonsproject.org/projects/waitress/en/stable/runner.html
    host = args.host,          
    port = args.port,          
    url_scheme = 'http',     
    threads = args.threads,    
    connection_limit = args.limit+1,
    max_request_header_size = HRsizes.tobytes(args.maxH),
    max_request_body_size = HRsizes.tobytes(args.maxB),
    
)
#<-------------------DO NOT WRITE ANY CODE AFTER THIS
end_time = datetime.datetime.now()
print('')
print('◉ stop server @ [{}]'.format(end_time))
print('◉ server up-time was [{}]'.format(end_time - start_time))


